@@ -3,6 +3,7 @@
 import os
 from itertools import chain
 import tempfile
+import logging
 import locale
 from tryton.common import \
         domain_inversion, eval_domain, localize_domain, \
@@ -14,8 +15,13 @@ import datetime
 import decimal
 from decimal import Decimal
 import math
+from pathlib import Path
 from tryton.common import RPCExecute, RPCException
-import tryton.rpc as rpc
+from tryton.common.htmltextbuffer import guess_decode
+from tryton.pyson import PYSONDecoder
+from tryton.config import CONFIG
+
+logger = logging.getLogger(__name__)
 
 
 class Field(object):
@@ -34,6 +40,7 @@ class Field(object):
     def __init__(self, attrs):
         self.attrs = attrs
         self.name = attrs['name']
+        self.views = set()
 
     def sig_changed(self, record):
         record.on_change([self.name])
@@ -59,13 +66,23 @@ class Field(object):
     def validation_domains(self, record, pre_validate=None):
         return concat(*self.domains_get(record, pre_validate))
 
-    def get_context(self, record):
-        context = record.get_context().copy()
-        if record.parent:
-            context.update(record.parent.get_context())
+    def get_context(self, record, record_context=None, local=False):
+        if record_context is not None:
+            context = record_context.copy()
+        else:
+            context = record.get_context(local=local)
         context.update(record.expr_eval(self.attrs.get('context', {})))
         return context
 
+    def get_search_context(self, record):
+        context = self.get_context(record)
+        context.update(record.expr_eval(self.attrs.get('search_context', {})))
+        return context
+
+    def get_search_order(self, record):
+        order = record.expr_eval(self.attrs.get('search_order', None))
+        return order
+
     def _is_empty(self, record):
         return not self.get_eval(record)
 
@@ -191,13 +208,49 @@ class CharField(Field):
     def get(self, record):
         return super(CharField, self).get(record) or self._default
 
+    def set_client(self, record, value, force_change=False):
+        if isinstance(value, bytes):
+            try:
+                value = guess_decode(value)
+            except UnicodeDecodeError:
+                logger.warning(
+                    "The encoding can not be guessed for field '%(name)s'",
+                    {'name': self.name})
+                value = None
+        super().set_client(record, value, force_change)
+
 
 class SelectionField(Field):
 
     _default = None
 
-    def get_client(self, record):
-        return record.value.get(self.name)
+
+class MultiSelectionField(Field):
+
+    _default = None
+
+    def get(self, record):
+        value = super().get(record)
+        if not value:
+            value = self._default
+        else:
+            value.sort()
+        return value
+
+    def get_eval(self, record):
+        value = super().get_eval(record)
+        if value is None:
+            value = []
+        return value
+
+    def set_client(self, record, value, force_change=False):
+        if value is None:
+            value = []
+        if isinstance(value, str):
+            value = [value]
+        if value:
+            value = sorted(value)
+        super().set_client(record, value, force_change=force_change)
 
 
 class DateTimeField(Field):
@@ -231,7 +284,7 @@ class DateTimeField(Field):
 
     def date_format(self, record):
         context = self.get_context(record)
-        return context.get('date_format', '%x')
+        return common.date_format(context.get('date_format'))
 
     def time_format(self, record):
         return record.expr_eval(self.attrs['format'])
@@ -250,7 +303,7 @@ class DateField(Field):
 
     def date_format(self, record):
         context = self.get_context(record)
-        return context.get('date_format', '%x')
+        return common.date_format(context.get('date_format'))
 
 
 class TimeField(Field):
@@ -277,19 +330,18 @@ class TimeDeltaField(Field):
     def _is_empty(self, record):
         return self.get(record) is None
 
-    def converter(self, record):
-        # TODO allow local context converter
-        return rpc.CONTEXT.get(self.attrs.get('converter'))
+    def converter(self, group):
+        return group.context.get(self.attrs.get('converter'))
 
     def set_client(self, record, value, force_change=False):
-        if isinstance(value, basestring):
-            value = common.timedelta.parse(value, self.converter(record))
+        if isinstance(value, str):
+            value = common.timedelta.parse(value, self.converter(record.group))
         super(TimeDeltaField, self).set_client(
             record, value, force_change=force_change)
 
     def get_client(self, record):
         value = super(TimeDeltaField, self).get_client(record)
-        return common.timedelta.format(value, self.converter(record))
+        return common.timedelta.format(value, self.converter(record.group))
 
 
 class FloatField(Field):
@@ -308,6 +360,28 @@ class FloatField(Field):
         shift = int(round(math.log(abs(factor), 10)))
         return (digits[0] + shift, digits[1] - shift)
 
+    def get_symbol(self, record, symbol):
+        if record and symbol in record.group.fields:
+            value = self.get(record) or 0
+            sign = 1
+            if value < 0:
+                sign = -1
+            elif value == 0:
+                sign = 0
+            symbol_field = record.group.fields[symbol]
+            symbol_name = symbol_field.attrs.get('relation')
+            symbol_id = symbol_field.get(record)
+            if symbol_name and symbol_id is not None and symbol_id >= 0:
+                try:
+                    return RPCExecute(
+                        'model', symbol_name, 'get_symbol', symbol_id, sign,
+                        context=record.get_context())
+                except RPCException:
+                    logger.warn(
+                        "Fail to fetch symbol for %s,%s",
+                        symbol_name, symbol_id)
+        return '', 1
+
     def convert(self, value):
         try:
             return locale.atof(value)
@@ -315,10 +389,14 @@ class FloatField(Field):
             return self._default
 
     def set_client(self, record, value, force_change=False, factor=1):
-        if isinstance(value, basestring):
+        if isinstance(value, str):
             value = self.convert(value)
         if value is not None:
-            value /= factor
+            # Keep the same type
+            if isinstance(value, int):
+                value //= factor
+            else:
+                value /= factor
         super(FloatField, self).set_client(record, value,
             force_change=force_change)
 
@@ -326,14 +404,14 @@ class FloatField(Field):
         value = record.value.get(self.name)
         if value is not None:
             digits = self.digits(record, factor=factor)
+            d = value * factor
+            if not isinstance(d, Decimal):
+                d = Decimal(repr(d))
             if digits:
-                p = digits[1]
+                p = int(digits[1])
             else:
-                d = value * factor
-                if not isinstance(d, Decimal):
-                    d = Decimal(repr(d))
                 p = -int(d.as_tuple().exponent)
-            return locale.format('%.*f', (p, value * factor), True)
+            return locale.localize('{0:.{1}f}'.format(d, p), True)
         else:
             return ''
 
@@ -342,7 +420,7 @@ class NumericField(FloatField):
 
     def convert(self, value):
         try:
-            return locale.atof(value, Decimal)
+            return Decimal(locale.delocalize(value))
         except decimal.InvalidOperation:
             return self._default
 
@@ -398,38 +476,42 @@ class M2OField(Field):
         return self.get(record) is None
 
     def get_client(self, record):
-        rec_name = record.value.get(self.name + '.rec_name')
+        rec_name = record.value.get(self.name + '.', {}).get('rec_name')
         if rec_name is None:
             self.set(record, self.get(record))
-            rec_name = record.value.get(self.name + '.rec_name') or ''
-        return rec_name
+            rec_name = record.value.get(self.name + '.', {}).get('rec_name')
+        return rec_name or ''
 
     def set_client(self, record, value, force_change=False):
         if isinstance(value, (tuple, list)):
             value, rec_name = value
         else:
             if value == self.get(record):
-                rec_name = record.value.get(self.name + '.rec_name', '')
+                rec_name = record.value.get(
+                    self.name + '.', {}).get('rec_name', '')
             else:
                 rec_name = ''
-        record.value[self.name + '.rec_name'] = rec_name
+        if value and value < 0 and self.name != record.parent_name:
+            value, rec_name = None, ''
+        record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name
         super(M2OField, self).set_client(record, value,
             force_change=force_change)
 
     def set(self, record, value):
-        rec_name = record.value.get(self.name + '.rec_name') or ''
-        if not rec_name and value >= 0:
+        rec_name = record.value.get(self.name + '.', {}).get('rec_name') or ''
+        if not rec_name and value is not None and value >= 0:
             try:
                 result, = RPCExecute('model', self.attrs['relation'], 'read',
                     [value], ['rec_name'])
             except RPCException:
                 return
             rec_name = result['rec_name'] or ''
-        record.value[self.name + '.rec_name'] = rec_name
+        record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name
         record.value[self.name] = value
 
-    def get_context(self, record):
-        context = super(M2OField, self).get_context(record)
+    def get_context(self, record, record_context=None, local=False):
+        context = super(M2OField, self).get_context(
+            record, record_context=record_context, local=local)
         if self.attrs.get('datetime_field'):
             context['_datetime'] = record.get_eval(
                 )[self.attrs.get('datetime_field')]
@@ -491,7 +573,7 @@ class O2MField(Field):
     def _set_default_value(self, record, fields=None):
         if record.value.get(self.name) is not None:
             return
-        from group import Group
+        from .group import Group
         parent_name = self.attrs.get('relation_field', '')
         fields = fields or {}
         context = record.expr_eval(self.attrs.get('context', {}))
@@ -531,11 +613,12 @@ class O2MField(Field):
             if record2 in record_removed or record2 in record_deleted:
                 continue
             if record2.id >= 0:
-                values = record2.get()
-                values.pop(parent_name, None)
-                if record2.modified and values:
-                    to_write.extend(([record2.id], values))
-                to_add.append(record2.id)
+                if record2.modified:
+                    values = record2.get()
+                    values.pop(parent_name, None)
+                    if values:
+                        to_write.extend(([record2.id], values))
+                    to_add.append(record2.id)
             else:
                 values = record2.get()
                 values.pop(parent_name, None)
@@ -556,9 +639,10 @@ class O2MField(Field):
         if record.value.get(self.name) is None:
             return {}
         result = {}
-        for record2 in (record.value[self.name]
-                + record.value[self.name].record_removed
-                + record.value[self.name].record_deleted):
+        record_modified = (r for r in record.value[self.name] if r.modified)
+        for record2 in chain(record_modified,
+                record.value[self.name].record_removed,
+                record.value[self.name].record_deleted):
             result.update(record2.get_timestamp())
         return result
 
@@ -581,12 +665,12 @@ class O2MField(Field):
                         skip={self.attrs.get('relation_field', '')}))
         return result
 
-    def _set_value(self, record, value, default=False):
+    def _set_value(self, record, value, default=False, modified=False):
         self._set_default_value(record)
         group = record.value[self.name]
         if value is None:
             value = []
-        if not value or isinstance(value[0], (int, long)):
+        if not value or isinstance(value[0], int):
             mode = 'list ids'
         else:
             mode = 'list values'
@@ -604,13 +688,19 @@ class O2MField(Field):
                 group.load_fields(fields)
 
         if mode == 'list ids':
-            for old_record in group:
-                if old_record.id not in value:
-                    group.remove(old_record, remove=True, signal=False)
-            group.load(value)
+            records_to_remove = [r for r in group if r.id not in value]
+            for record_to_remove in records_to_remove:
+                group.remove(record_to_remove, remove=True, signal=False)
+            group.load(value, modified=modified or default)
         else:
             for vals in value:
-                new_record = record.value[self.name].new(default=False)
+                if 'id' in vals:
+                    new_record = group.get(vals['id'])
+                    if not new_record:
+                        new_record = group.new(
+                            default=False, obj_id=vals['id'])
+                else:
+                    new_record = group.new(default=False)
                 if default:
                     # Don't validate as parent will validate
                     new_record.set_default(vals, signal=False, validate=False)
@@ -633,7 +723,7 @@ class O2MField(Field):
             fields = record.group.fields
         if fields:
             fields = dict((fname, field.attrs)
-                for fname, field in fields.iteritems())
+                for fname, field in fields.items())
 
         record.value[self.name] = None
         self._set_default_value(record, fields=fields)
@@ -648,13 +738,14 @@ class O2MField(Field):
         if value is None:
             value = []
         # domain inversion could try to set id as value
-        if isinstance(value, (int, long)):
+        if isinstance(value, int):
             value = [value]
 
         previous_ids = self.get_eval(record)
-        self._set_value(record, value)
         # The order of the ids is not significant
-        if set(previous_ids) != set(value):
+        modified = set(previous_ids) != set(value)
+        self._set_value(record, value, modified=modified)
+        if modified:
             record.modified_fields.setdefault(self.name)
             record.signal('record-modified')
             self.sig_changed(record)
@@ -673,7 +764,7 @@ class O2MField(Field):
         record.modified_fields.setdefault(self.name)
         self._set_default_value(record)
         if isinstance(value, (list, tuple)):
-            self._set_value(record, value)
+            self._set_value(record, value, modified=True)
             return
 
         if value and (value.get('add') or value.get('update')):
@@ -682,40 +773,59 @@ class O2MField(Field):
             values = chain(value.get('update', []),
                 (d for _, d in value.get('add', [])))
             field_names = set(f for v in values
-                for f in v if f not in fields and f != 'id')
+                for f in v if f not in fields and f != 'id' and '.' not in f)
             if field_names:
                 try:
-                    fields = RPCExecute('model', self.attrs['relation'],
+                    new_fields = RPCExecute('model', self.attrs['relation'],
                         'fields_get', list(field_names), context=context)
                 except RPCException:
                     return
             else:
-                fields = {}
+                new_fields = {}
 
-        to_remove = []
-        for record2 in record.value[self.name]:
-            if not record2.id:
-                to_remove.append(record2)
+        group = record.value[self.name]
+        if value and value.get('delete'):
+            for record_id in value['delete']:
+                record2 = group.get(record_id)
+                if record2 is not None:
+                    group.remove(
+                        record2, remove=False, signal=False,
+                        force_remove=False)
         if value and value.get('remove'):
             for record_id in value['remove']:
-                record2 = record.value[self.name].get(record_id)
+                record2 = group.get(record_id)
                 if record2 is not None:
-                    to_remove.append(record2)
-        for record2 in to_remove:
-            record.value[self.name].remove(record2, signal=False,
-                force_remove=False)
+                    group.remove(
+                        record2, remove=True, signal=False,
+                        force_remove=False)
 
         if value and (value.get('add') or value.get('update', [])):
-            record.value[self.name].add_fields(fields)
+            # First set already added fields to prevent triggering a
+            # second on_change call
+            for vals in value.get('update', []):
+                if 'id' not in vals:
+                    continue
+                record2 = group.get(vals['id'])
+                if record2 is not None:
+                    vals_to_set = {
+                        k: v for k, v in vals.items() if k not in new_fields}
+                    record2.set_on_change(vals_to_set)
+
+            record.value[self.name].add_fields(new_fields)
             for index, vals in value.get('add', []):
-                new_record = record.value[self.name].new(default=False)
-                record.value[self.name].add(new_record, index, signal=False)
+                new_record = None
+                id_ = vals.pop('id', None)
+                if id_ is not None:
+                    new_record = group.get(id_)
+                if not new_record:
+                    new_record = group.new(obj_id=id_, default=False)
+                group.add(new_record, index, signal=False)
                 new_record.set_on_change(vals)
 
             for vals in value.get('update', []):
                 if 'id' not in vals:
                     continue
-                record2 = record.value[self.name].get(vals['id'])
+                record2 = group.get(vals['id'])
                 if record2 is not None:
                     record2.set_on_change(vals)
 
@@ -770,10 +880,17 @@ class ReferenceField(Field):
 
     _default = None
 
+    def _is_empty(self, record):
+        result = super(ReferenceField, self)._is_empty(record)
+        if not result and (record.value[self.name] is None
+                or record.value[self.name][1] < 0):
+            result = True
+        return result
+
     def get_client(self, record):
         if record.value.get(self.name):
             model, _ = record.value[self.name]
-            name = record.value.get(self.name + '.rec_name') or ''
+            name = record.value.get(self.name + '.', {}).get('rec_name') or ''
             return model, name
         else:
             return None
@@ -781,13 +898,14 @@ class ReferenceField(Field):
     def get(self, record):
         if (record.value.get(self.name)
                 and record.value[self.name][0]
+                and record.value[self.name][1] is not None
                 and record.value[self.name][1] >= -1):
             return ','.join(map(str, record.value[self.name]))
         return None
 
     def set_client(self, record, value, force_change=False):
         if value:
-            if isinstance(value, basestring):
+            if isinstance(value, str):
                 value = value.split(',')
             ref_model, ref_id = value
             if isinstance(ref_id, (tuple, list)):
@@ -799,10 +917,11 @@ class ReferenceField(Field):
                     except ValueError:
                         pass
                 if '%s,%s' % (ref_model, ref_id) == self.get(record):
-                    rec_name = record.value.get(self.name + '.rec_name', '')
+                    rec_name = record.value.get(
+                        self.name + '.', {}).get('rec_name') or ''
                 else:
                     rec_name = ''
-            record.value[self.name + '.rec_name'] = rec_name
+            record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name
             value = (ref_model, ref_id)
         super(ReferenceField, self).set_client(record, value,
             force_change=force_change)
@@ -811,7 +930,7 @@ class ReferenceField(Field):
         if not value:
             record.value[self.name] = self._default
             return
-        if isinstance(value, basestring):
+        if isinstance(value, str):
             ref_model, ref_id = value.split(',')
             if not ref_id:
                 ref_id = None
@@ -822,8 +941,8 @@ class ReferenceField(Field):
                     pass
         else:
             ref_model, ref_id = value
-        rec_name = record.value.get(self.name + '.rec_name') or ''
-        if ref_model and ref_id >= 0:
+        rec_name = record.value.get(self.name + '.', {}).get('rec_name') or ''
+        if ref_model and ref_id is not None and ref_id >= 0:
             if not rec_name and ref_id >= 0:
                 try:
                     result, = RPCExecute('model', ref_model, 'read', [ref_id],
@@ -834,12 +953,13 @@ class ReferenceField(Field):
         elif ref_model:
             rec_name = ''
         else:
-            rec_name = str(ref_id)
+            rec_name = str(ref_id) if ref_id is not None else ''
         record.value[self.name] = ref_model, ref_id
-        record.value[self.name + '.rec_name'] = rec_name
+        record.value.setdefault(self.name + '.', {})['rec_name'] = rec_name
 
-    def get_context(self, record):
-        context = super(ReferenceField, self).get_context(record)
+    def get_context(self, record, record_context=None, local=False):
+        context = super(ReferenceField, self).get_context(
+            record, record_context, local=local)
         if self.attrs.get('datetime_field'):
             context['_datetime'] = record.get_eval(
                 )[self.attrs.get('datetime_field')]
@@ -861,33 +981,69 @@ class ReferenceField(Field):
         else:
             model = None
         screen_domain, attr_domain = self.domains_get(record)
+        screen_domain = filter_leaf(screen_domain, self.name, model)
         screen_domain = prepare_reference_domain(screen_domain, self.name)
         return concat(localize_domain(
-                filter_leaf(screen_domain, self.name, model),
-                strip_target=True), attr_domain)
+                screen_domain, self.name, strip_target=True), attr_domain)
 
     def get_models(self, record):
         screen_domain, attr_domain = self.domains_get(record)
+        screen_domain = prepare_reference_domain(screen_domain, self.name)
         return extract_reference_models(
             concat(screen_domain, attr_domain), self.name)
 
 
 class _FileCache(object):
-    def __init__(self, path):
-        self.path = path
+    def __init__(self, data=None):
+        _, filename = tempfile.mkstemp(prefix='tryton_')
+        self.path = Path(filename)
+        self.suffixes = {}
+        if data:
+            with open(self.path, 'wb') as fp:
+                fp.write(data)
+
+    @property
+    def data(self):
+        with open(self.path, 'rb') as fp:
+            return fp.read()
+
+    def __del__(self):
+        try:
+            os.remove(self.path)
+        except IOError:
+            pass
+        for path in self.suffixes.values():
+            try:
+                os.remove(path)
+            except IOError:
+                pass
+
+    def with_suffix(self, suffix):
+        if suffix in self.suffixes:
+            return self.suffixes[suffix]
+        _, filename = tempfile.mkstemp(prefix='tryton_', suffix=suffix)
+        self.suffixes[suffix] = path = Path(filename)
+        with open(path, 'wb') as fp:
+            fp.write(self.data)
+        return path
 
 
 class BinaryField(Field):
 
     _default = None
-    cast = bytearray if bytes == str else bytes
+
+    def _set_file_cache(self, record, data):
+        if isinstance(data, str):
+            data = data.encode('utf-8')
+        file_cache = _FileCache(data)
+        self.set(record, file_cache)
+        return file_cache
 
     def get(self, record):
         result = record.value.get(self.name, self._default)
         if isinstance(result, _FileCache):
             try:
-                with open(result.path, 'rb') as fp:
-                    result = self.cast(fp.read())
+                result = result.data
             except IOError:
                 result = self.get_data(record)
         return result
@@ -896,10 +1052,7 @@ class BinaryField(Field):
         return self.get(record)
 
     def set_client(self, record, value, force_change=False):
-        _, filename = tempfile.mkstemp(prefix='tryton_')
-        with open(filename, 'wb') as fp:
-            fp.write(value or '')
-        self.set(record, _FileCache(filename))
+        self._set_file_cache(record, value or b'')
         record.modified_fields.setdefault(self.name)
         record.signal('record-modified')
         self.sig_changed(record)
@@ -910,37 +1063,49 @@ class BinaryField(Field):
         result = record.value.get(self.name) or 0
         if isinstance(result, _FileCache):
             result = os.stat(result.path).st_size
-        elif isinstance(result, (basestring, bytes, bytearray)):
+        elif isinstance(result, (str, bytes)):
             result = len(result)
         return result
 
     def get_data(self, record):
         if not isinstance(record.value.get(self.name),
-                (basestring, bytes, bytearray, _FileCache)):
+                (str, bytes, _FileCache)):
             if record.id < 0:
-                return ''
+                return b''
             context = record.get_context()
             try:
                 values, = RPCExecute('model', record.model_name, 'read',
                     [record.id], [self.name], context=context)
             except RPCException:
-                return ''
-            _, filename = tempfile.mkstemp(prefix='tryton_')
-            with open(filename, 'wb') as fp:
-                fp.write(values[self.name] or '')
-            self.set(record, _FileCache(filename))
+                return b''
+            self._set_file_cache(record, values[self.name] or b'')
         return self.get(record)
 
+    def get_filename(self, record, suffix=None):
+        data = self.get_data(record)
+        file_cache = record.value.get(self.name)
+        if not isinstance(file_cache, _FileCache):
+            file_cache = self._set_file_cache(record, data)
+        if suffix:
+            filename = file_cache.with_suffix(suffix)
+        else:
+            filename = file_cache.path
+        return filename
+
 
 class DictField(Field):
 
     _default = {}
 
+    def __init__(self, attrs):
+        super(DictField, self).__init__(attrs)
+        self.keys = {}
+
     def get(self, record):
-        return super(DictField, self).get(record) or self._default
+        return (super().get(record) or self._default).copy()
 
     def get_client(self, record):
-        return super(DictField, self).get_client(record) or self._default
+        return super().get_client(record)
 
     def validation_domains(self, record, pre_validate=None):
         screen_domain, attr_domain = self.domains_get(record, pre_validate)
@@ -952,11 +1117,76 @@ class DictField(Field):
 
     def date_format(self, record):
         context = self.get_context(record)
-        return context.get('date_format', '%x')
+        return common.date_format(context.get('date_format'))
 
     def time_format(self, record):
         return '%X'
 
+    def add_keys(self, keys, record):
+        schema_model = self.attrs['schema_model']
+        context = self.get_context(record)
+        domain = self.domain_get(record)
+        batchlen = min(10, CONFIG['client.limit'])
+        for i in range(0, len(keys), batchlen):
+            sub_keys = keys[i:i + batchlen]
+            try:
+                key_ids = RPCExecute('model', schema_model, 'search',
+                    [('name', 'in', sub_keys), domain], 0,
+                    CONFIG['client.limit'], None, context=context)
+            except RPCException:
+                key_ids = []
+            if not key_ids:
+                continue
+            try:
+                values = RPCExecute('model', schema_model,
+                    'get_keys', key_ids, context=context)
+            except RPCException:
+                values = []
+            if not values:
+                continue
+            self.keys.update({k['name']: k for k in values})
+
+    def add_new_keys(self, key_ids, record):
+        schema_model = self.attrs['schema_model']
+        context = self.get_context(record)
+        try:
+            new_fields = RPCExecute('model', schema_model,
+                'get_keys', key_ids, context=context)
+        except RPCException:
+            new_fields = []
+
+        new_keys = []
+        for new_field in new_fields:
+            name = new_field['name']
+            new_keys.append(name)
+            self.keys[name] = new_field
+
+        return new_keys
+
+    def validate(self, record, softvalidation=False, pre_validate=None):
+        valid = super(DictField, self).validate(
+            record, softvalidation, pre_validate)
+
+        if self.attrs.get('readonly'):
+            return valid
+
+        decoder = PYSONDecoder()
+        field_value = self.get_eval(record)
+        domain = []
+        for key in field_value:
+            if key not in self.keys:
+                continue
+            key_domain = self.keys[key].get('domain')
+            if key_domain:
+                domain.append(decoder.decode(key_domain))
+
+        valid_value = eval_domain(domain, field_value)
+        if not valid_value:
+            self.get_state_attrs(record)['invalid'] = 'domain'
+
+        return valid and valid_value
+
+
 TYPES = {
     'char': CharField,
     'integer': IntegerField,
@@ -968,10 +1198,12 @@ TYPES = {
     'one2many': O2MField,
     'reference': ReferenceField,
     'selection': SelectionField,
+    'multiselection': MultiSelectionField,
     'boolean': BooleanField,
     'datetime': DateTimeField,
     'date': DateField,
     'time': TimeField,
+    'timestamp': DateTimeField,
     'timedelta': TimeDeltaField,
     'one2one': O2OField,
     'binary': BinaryField,
